#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

import logging
from threading import Thread
from typing import Dict, List, NamedTuple

from transaction.interfaces import IDataManager
from zope.interface import implementer

from workspaces.notification.services.notification_service import (
    NotificationServiceIF,
    NotificationServiceRESTClient,
)

logger = logging.getLogger("capability.utility")
NOTIF_SETTINGS_KEY = "edu.nrao.workspaces.NotificationSettings"


class PendingNotification(NamedTuple):
    """
    Reifies a notification to be sent.
    """

    notification: str
    arguments: Dict


@implementer(IDataManager)
class TransactionalNotificationSender(NotificationServiceIF):
    """
    Implements the NotificationServiceIF in a way that ties into the Zope Transaction manager.

    This class has a "front-end" and a "back-end". The front-end is the NotificationServiceIF, which
    is currently many, many different notification methods, but should soon be resolved down to just one,
    :meth:`send_email`.

    The back-end is the suite of methods mandated by the IDataManager interface, which has a variety of
    things, but is the various methods with `tpc` in the name.
    """

    def __init__(self, transaction_manager):
        logger.debug("Constructing TransactionalNotificationSender")
        # join the transaction
        self.transaction_manager = transaction_manager
        self.join_transaction()

        # we don't begin life in a transaction
        self.in_commit = False

        # let's have a list of notifications to send when we commi
        self.notifications: List[PendingNotification] = []

    def join_transaction(self):
        self.transaction_manager.get().join(self)

    def send_email(self, notification: str, arguments: Dict):
        """
        Sends an email.

        In reality, arranges for an email to be sent when the transaction is committed.

        :param notification:  the notification to send
        :param arguments:     the JSON to attach to it
        """
        logger.debug("Stashing a notification to %s for sending later", notification)
        self.notifications.append(PendingNotification(notification, arguments))

    def start_sending_notifications(self):
        if self.notifications:
            logger.debug("Setting up notification sending thread")
            sender_thread = Thread(
                target=self._send_prepared_notifications,
                args=(self.notifications[:],),
                name="capability notification sender",
            )
            sender_thread.start()
        else:
            logger.debug("No pending notifications to send")

    @staticmethod
    def _send_prepared_notifications(notifications):
        try:
            logger.debug("Sending %s previously prepared notifications", len(notifications))
            client = NotificationServiceRESTClient()
            for notification in notifications:
                logger.info("My arguments are %s", notification.arguments)
                client.send_email(notification.notification, notification.arguments)
        except Exception:
            logger.exception("Unable to send notifications")

    # Transaction machinery
    def _cleanup(self):
        self.notifications = []

    def tpc_begin(self, transaction):
        logger.debug("Beginning transaction")
        pass

    def commit(self, transaction):
        logger.debug("Committing transaction")
        self.in_commit = True

    def tpc_vote(self, transaction):
        logger.debug("Voting on transaction")
        # all we have to do to cancel the transaction is throw an exception here
        pass

    def tpc_finish(self, transaction):
        logger.debug("Finishing transaction")
        self.start_sending_notifications()
        self.in_commit = False
        self._cleanup()

    def tpc_abort(self, transaction):
        logger.debug("Aborting transaction")
        self.in_commit = False
        self._cleanup()

    """
    IDataManager defines two abort methods, one for two-phase commit (tpc_abort) and one for use outside of a two-phase
    commit. However, we do not make a distinction between the two aborts behaviorally, so this creates a synonym from
    one to the other.
    """
    abort = tpc_abort

    # this is mandatory in the Zope Transaction API but I don't know what it does
    def sortKey(self):
        return "safety first"
