#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Coverage audit proof-of-concept masquerading as tests """
import os
import shutil
import tempfile
from pathlib import Path
from unittest.mock import patch

import coverage_audit

# pylint: disable=E0401, R1721, W0621
import pytest
from coverage_audit import (
    COVERAGE_REPORT_XML,
    CoverageReportGrabber,
    CoverageReportParser,
    Reporter,
    find_project_root,
)

RUN_ALL = True


@pytest.fixture(scope="module")
def project_root() -> Path:
    """
    Get the project root once and use it in all tests

    :return: top level under which audit targets will be found
    """
    return find_project_root()


@pytest.fixture(scope="module")
def test_coverage_report() -> Path:
    """
    Find the testing copy of the gitlab coverage report.

    :return: path to report file
    """
    cur_dir = Path.cwd()
    xmls = [file for file in cur_dir.parent.rglob(COVERAGE_REPORT_XML)]
    for xml in xmls:
        if "coverage_audit/test" in str(xml):
            return xml


def test_main_launches_cov_audit(capsys):
    """
    Is coverage_audit.main() functional?

    :return:
    """

    tmpdir = Path(tempfile.mkdtemp())

    orig_dir = Path.cwd()

    # try it in current directory first
    os.chdir(tmpdir)
    coverage_audit.main()
    cov_rpt_file = tmpdir / COVERAGE_REPORT_XML
    assert cov_rpt_file.exists()
    result = capsys.readouterr().out
    lines = result.strip().split("\n")
    num_lines = len(lines)
    assert num_lines > 1

    # go back to where we were, then pass in the destination
    os.chdir(orig_dir)
    cov_rpt_file.unlink()
    coverage_audit.main([str(tmpdir)])
    assert cov_rpt_file.exists()
    result = capsys.readouterr().out
    lines = result.strip().split("\n")
    assert len(lines) == num_lines

    shutil.rmtree(tmpdir)


@pytest.mark.skipif(not RUN_ALL, reason="temporary skip")
def test_expected_targets_created(test_coverage_report):
    """
    Do the audit targets in the testing coverage.xml have the metadata we expect?

    :return:
    """
    crp = CoverageReportParser(test_coverage_report)

    crp.parse()
    # we're expecting this many "classes" to have coverage < 100%
    assert len(crp.targets) == 39
    for target in crp.targets:
        assert target.branch_rate >= 0.0
        assert target.line_rate > 0.0


@pytest.mark.skipif(not RUN_ALL, reason="temporary skip")
def test_all_targets_analyzed(test_coverage_report):
    """
    Have we dug into every target to get its 411?

    :return:
    """
    crp = CoverageReportParser(test_coverage_report)

    with patch("coverage_audit.ElementDigger.dig") as mock:
        crp.parse()

    assert mock.call_count == len(crp.targets)


@pytest.mark.skipif(not RUN_ALL, reason="temporary skip")
def test_parses_target_accurately(test_coverage_report):
    """
    Do we find the attributes we expect in a given target created from
    an element of the coverage report?

    :return:
    """
    crp = CoverageReportParser(test_coverage_report)
    crp.parse()

    module_found = False
    for target in crp.targets:
        if target.path.name.endswith("delivery.py"):
            assert target.line_rate == 0.8077
            assert target.branch_rate == 0.5
            assert len(target.modules) == 2
            for module in target.modules:
                assert module.path.name.endswith("delivery.py")
            module_found = True
            break

    assert module_found


@pytest.mark.skipif(not RUN_ALL, reason="temporary skip")
def test_creates_expected_digest(test_coverage_report: Path, capsys):
    """
    For the test copy of coverage.xml, do we get the output we should?

    :return:
    """
    parser = CoverageReportParser(test_coverage_report)
    parser.parse()
    reporter = Reporter(parser.targets)
    reporter.get_reports()
    result = capsys.readouterr().out
    lines = result.strip().split("\n")
    assert len(lines) == len(parser.targets) + 1


@pytest.mark.skipif(not RUN_ALL, reason="temporary skip")
def test_fetches_latest_coverage_report():
    """
    Can we retrieve the latest coverage report from gitlab with an http request?

    :return:
    """
    tmpdir = Path(tempfile.mkdtemp())
    latest_coverage_report = CoverageReportGrabber(tmpdir).grab()
    assert latest_coverage_report.exists()

    shutil.rmtree(tmpdir)


@pytest.mark.skipif(not RUN_ALL, reason="temporary skip")
def test_parses_latest_coverage_report(test_coverage_report: Path):
    """
    Do we get plausible results from the latest gitlab coverage report
    (and are we indeed parsing the real one, not the test copy)?

    :return:
    """
    tmpdir = Path(tempfile.mkdtemp())
    latest_report = CoverageReportGrabber(tmpdir).grab()
    parser = CoverageReportParser(latest_report)
    parser.parse()

    results = parser.targets
    assert len(results) > 0

    # results should be different from those of parsing our test copy
    parser = CoverageReportParser(test_coverage_report)
    parser.parse()
    test_results = parser.targets

    # if number of results is same, be sure at least one of the results is different;
    # otherwise, we can be confident we're reading the real coverage report....
    diff_found = False
    if len(results) == len(test_results):
        for result in results:
            if result not in test_results:
                diff_found = True
                break
    else:
        diff_found = True
    assert diff_found

    # ... but let's confirm that: make the lists equal length and compare results
    len_diff = abs(len(test_results) - len(results))
    if len(results) > len(test_results):
        results = results[: len(results) - len_diff]
    else:
        test_results = results[: len(test_results) - len_diff]

    diff_found = False
    for result in results:
        if result not in test_results:
            diff_found = True
            break
    assert diff_found

    shutil.rmtree(tmpdir)
